from peliqan.exceptions import PeliqanClientException, OperationNotSupported
from peliqan.client.base import BaseClient
from peliqan.utils import empty, _check_record_byte_size

import pandas as pd
import json


class DBClient(BaseClient):
    is_trino = False

    def __init__(self, connection, jwt, backend_url):
        super(DBClient, self).__init__(jwt, backend_url)
        self.connection = connection

    def db_via_proxy(self, db_name, schema_name, table_name, pk, action, **kwargs):
        # pk has same identity as empty declared above
        if pk is empty:
            raise PeliqanClientException("'pk' must be provided.")

        payload = {
            "connection": self.connection,
            "dbName": db_name,
            "schemaName": schema_name,
            "tableName": table_name,
            "pk": pk,
            "action": action,
            "kwargs": kwargs
        }

        url = f"{self.BACKEND_URL}/api/proxy/db/"
        return self.call_backend('post', url, json=payload)

    def create_schema(self, db_name, schema_name):
        response = self.db_via_proxy(db_name, schema_name, None, None, 'create_schema')
        return response

    def create_table(self, db_name, schema_name, table_name, fields=None, pk=None):
        response = self.db_via_proxy(db_name, schema_name, table_name, pk, 'create_table', fields=fields)
        return response

    def insert(self, db_name, schema_name, table_name, *args, **kwargs):
        # allow using both keyword arguments or a dict as argument:
        # pq.insert("my_db", "my_table", name='John', city='NY') or pq.insert("my_db", "my_table", contact_obj)
        kwargs = self.args_to_kwargs(args, kwargs)
        response = self.db_via_proxy(db_name, schema_name, table_name, None, 'insert', **kwargs)
        return response

    def update(self, db_name, schema_name, table_name, pk=empty, *args, **kwargs):
        # allow using both keyword arguments or a dict as argument:
        # pq.update("my_db", "my_table", id=1, name='John', city='NY') or pq.update("my_db", "my_table", contact_obj)
        kwargs = self.args_to_kwargs(args, kwargs)
        response = self.db_via_proxy(db_name, schema_name, table_name, pk, 'update', **kwargs)
        return response

    def upsert(self, db_name, schema_name, table_name, pk=empty, *args, **kwargs):
        # allow using both keyword arguments or a dict as argument:
        # pq.update("my_db", "my_table", id=1, name='John', city='NY') or pq.update("my_db", "my_table", contact_obj)
        kwargs = self.args_to_kwargs(args, kwargs)
        response = self.db_via_proxy(db_name, schema_name, table_name, pk, 'upsert', **kwargs)
        return response

    def _to_dtype(self, column_type, tz, enable_datetime_as_string):
        if (
            column_type[0:6].lower() == "double" or
            column_type[0:5].lower() == "float" or
            column_type[0:7].lower() == "decimal" or
            column_type[0:7].lower() == "numeric"
        ):
            type_name = pd.Float64Dtype()

        elif (
            column_type[0:3].lower() == "int" or
            column_type[0:6].lower() == "bigint" or
            column_type[0:3].lower() == "num"
        ):  # todo add more field types
            type_name = pd.Int64Dtype()

        elif column_type[0:4].lower() == "bool":
            type_name = pd.BooleanDtype()

        elif not enable_datetime_as_string and (
            column_type == "date" or
            column_type[0:9].lower() == "timestamp" or
            column_type[0:8].lower() == "datetime"
        ):
            type_name = pd.DatetimeTZDtype(tz=tz)

        elif column_type[0:4].lower() == "json" and self.is_trino:
            type_name = 'json_str'

        elif column_type[0:4].lower() == "json":
            type_name = 'json'

        else:
            type_name = pd.StringDtype()

        return type_name

    def _convert_result_to_df(self, json_response, error, df=False, fillna_with=None, fillnat_with=None,
                              enable_python_types=True, enable_datetime_as_string=True, tz='UTC'):
        if error:
            raise PeliqanClientException(f"Client encountered a error while fetching records.\n{error}")

        records = json_response['records']
        column_data = json_response['columns']

        columns = []
        dtypes = []
        for column in column_data:
            dtypes.append((column[0], self._to_dtype(column[1], tz, enable_datetime_as_string)))
            columns.append(column[0])

        # Create dataframe
        dataframe = pd.DataFrame(records, columns=columns)
        if enable_python_types:
            for dtype in dtypes:
                if dtype[1] == 'json_str':
                    try:
                        dataframe[dtype[0]] = dataframe[dtype[0]].apply(json.loads)
                    except Exception:
                        pass

                elif dtype[1] != 'json':
                    dataframe[dtype[0]] = dataframe[dtype[0]].astype(dtype[1])

        dataframe.replace({pd.NaT: fillnat_with, pd.NA: fillna_with}, inplace=True)

        if df:
            return dataframe

        else:
            return dataframe.to_dict('records')

    def fetch(
        self, db_name, schema_name=None, table_name=None, query=None, df=False, fillna_with=None,
        fillnat_with=None, enable_python_types=True, enable_datetime_as_string=True, tz='UTC'
    ):
        json_response = self.db_via_proxy(db_name, schema_name, table_name, None, 'fetch', query=query)
        detail = json_response['detail']
        error = json_response['status'] == 'failed'
        if error:
            error = detail or error
        return self._convert_result_to_df(detail, error, df, fillna_with, fillnat_with, enable_python_types,
                                          enable_datetime_as_string, tz)

    def execute(self, db_name, query):
        kwargs = {'query': query}
        response = self.db_via_proxy(db_name, None, None, None, 'execute', **kwargs)
        return response

    def write(
        self,
        schema_name,
        table_name,
        records,
        object_schema=None,
        pk=None,
        db_name=None,
        transformer_mode='lossless',
        decimal_separator='.'
    ):
        if pk is None:
            pk = []

        if isinstance(records, pd.DataFrame):
            records.replace({pd.NaT: None, pd.NA: None}, inplace=True)
            records = records.to_dict('records')

        if not isinstance(records, (list, tuple)):
            if isinstance(records, dict):
                records = [records]
            else:
                raise PeliqanClientException("records has to be either a list or a dict")

        if not isinstance(pk, list):
            if isinstance(pk, str):
                pk = [pk]
            else:
                raise PeliqanClientException("pk has to be either a string or a list")

        _check_record_byte_size(records)
        kwargs = {
            'records': records,
            'object_schema': object_schema,
            'transformer_mode': transformer_mode,
            'decimal_separator': decimal_separator
        }
        response = self.db_via_proxy(db_name, schema_name, table_name, pk, 'write_records', **kwargs)
        return response


class PeliqanTrinoDBClient(DBClient):
    is_trino = True

    def create_schema(self, db_name, schema_name):
        raise OperationNotSupported("'create_schema' operation is not support by Peliqan Trino")

    def create_table(self, db_name, schema_name, table_name, fields=None, pk=None):
        raise OperationNotSupported("'create_table' operation is not support by Peliqan Trino")

    def insert(self, db_name, schema_name, table_name, *args, **kwargs):
        raise OperationNotSupported("'insert' operation is not support by Peliqan Trino")

    def update(self, db_name, schema_name, table_name, pk=empty, *args, **kwargs):
        raise OperationNotSupported("'update' operation is not support by Peliqan Trino")

    def upsert(self, db_name, schema_name, table_name, pk=empty, *args, **kwargs):
        raise OperationNotSupported("'update' operation is not support by Peliqan Trino")

    def write(
        self,
        schema_name,
        table_name,
        records,
        object_schema=None,
        pk=None,
        db_name=None,
        transformer_mode='lossless',
        decimal_separator='.'
    ):
        raise OperationNotSupported("'write' operation is not support by Peliqan Trino")

    def execute(self, db_name, query):
        raise OperationNotSupported("'execute' operation is not support by Peliqan Trino")

    def fetch(
        self, db_name=None, schema_name=None, table_name=None,
        query=None, table_id=None, df=False,
        fillna_with=None, fillnat_with=None,
        enable_python_types=True, enable_datetime_as_string=True, tz='UTC'
    ):
        """
            Return the records in a table.
            :param db_name: The name of a database.
            :param schema_name: The name of a schema a table belongs to.
            :param table_name: The name of a table.
            :param query: A valid Trino sql query.
            :param table_id: An integer value that uniquely identifies a table.
            :param df: This flag decides what should be returned by this function. df=True returns a dataframe. df=False returns an array of dicts, where each item represents a row with column name as the key.
            :param fillna_with: Replace empty/invalid values (NA) with a default value. This will also replace NaN values.
            :param fillnat_with: Replace invalid time/datetime values (NaT) with a default value.
            :param enable_python_types: This flag will cause the record's values to be returned as python types
            :param enable_datetime_as_string: This flag will cause date/time/datetime objects to be converted to string.
            :param tz: This is the timezone that will be used to convert date and datetime strings
            to timezone aware datetime objects. This value defaults to 'UTC'.

            :return:
        """

        # table_id will return the raw table data.
        # This causes multiple select and single select to be returned as json objects

        # query and table_name will return the data as if the queried table is a prepended statement.
        # This causes multiple select and single select to be returned as strings instead of a json object

        if query:
            url_suffix = f"?query={query}"

        elif table_id:
            url_suffix = f"?table_id={table_id}"

        elif table_name:
            url_suffix = f"?table_name={table_name}"

        else:
            raise PeliqanClientException("Input 'table_name' (table or view name), 'query' or 'table_id' must be set.")

        if db_name:
            url_suffix += f"&db_name={db_name}"

        if schema_name:
            url_suffix += f"&schema_name={schema_name}"

        # Fetch Trino catalog and SQL query with all CDC changes applied
        url = f"{self.BACKEND_URL}/api/database/resource/trino/records/" + url_suffix
        response_dict = self.call_backend("get", url)
        json_response = response_dict

        error = json_response['error']
        return self._convert_result_to_df(json_response, error, df, fillna_with, fillnat_with, enable_python_types,
                                          enable_datetime_as_string, tz)



from peliqan.client.base import BaseClient


class SFTPClient(BaseClient):
    def __init__(self, connection, jwt, backend_url):
        super(SFTPClient, self).__init__(jwt, backend_url)
        self.connection = connection

    def sftp_via_proxy(self, path, action, **kwargs):
        payload = {
            "connection": self.connection,
            "path": path,
            "action": action,
            "kwargs": kwargs
        }

        url = f"{self.BACKEND_URL}/api/proxy/sftp/"
        return self.call_backend('post', url, json=payload)

    def read_file(self, path, *args, **kwargs):
        kwargs = self.args_to_kwargs(args, kwargs)
        response = self.sftp_via_proxy(path, 'read_file', **kwargs)
        return response
    
    def dir(self, path, *args, **kwargs):
        kwargs = self.args_to_kwargs(args, kwargs)
        response = self.sftp_via_proxy(path, 'dir', **kwargs)
        return response




from json import JSONEncoder

from peliqan.exceptions import PeliqanJsonSerializerException, PeliqanClientException


class Empty:
    pass


empty = Empty()


def _serialize_dict(obj):
    for k in obj:
        obj[k] = _serialize_data(obj[k])

    return obj


def _serialize_array(obj):
    obj_len = len(obj)
    for i in range(obj_len):
        obj[i] = _serialize_data(obj[i])

    return obj


def _serialize_primitives(obj):
    formatted_obj = obj
    if isinstance(obj, float) and str(obj) == 'nan':
        formatted_obj = None

    return formatted_obj


def _serialize_others(obj):
    try:
        formatted_obj = JSONEncoder().encode(obj)
    except Exception as e:
        try:
            formatted_obj = str(obj)
        except Exception:
            raise PeliqanJsonSerializerException(
                f"Could not serialize {obj.__class__.__name__} with value {obj}. "
                f"Original error is {e}"
            )

    return formatted_obj


def _serialize_data(obj):
    if isinstance(obj, dict):
        formatted_obj = _serialize_dict(obj)

    elif type(obj) in (list, tuple):
        formatted_obj = _serialize_array(obj)

    elif isinstance(obj, (int, float, str)):
        formatted_obj = _serialize_primitives(obj)

    elif isinstance(obj, type(None)):
        formatted_obj = None

    else:
        formatted_obj = _serialize_others(obj)

    return formatted_obj


ONE_MB = 1024 * 1024
MAX_BATCH_BYTE_SIZE = ONE_MB * 10  # MB


def _check_record_byte_size(records):
    bytes_used = len(repr(records).encode('utf-8'))
    if bytes_used > MAX_BATCH_BYTE_SIZE:
        raise PeliqanClientException(
            f"Total size of all records, {int(bytes_used / ONE_MB)}MB, "
            f"exceeds max allowed size of {int(MAX_BATCH_BYTE_SIZE / ONE_MB)}MB."
        )



from peliqan.client.base import BaseClient
from peliqan.exceptions import PeliqanClientException


class WritebackClient(BaseClient):
    def __init__(self, connection, jwt, backend_url):
        super(WritebackClient, self).__init__(jwt, backend_url)
        self.connection = connection

    def request_endpoint_via_proxy(self, object_name, action, **kwargs):
        # send to proxy: self.connection, objectName, action, kwargs
        payload = {
            "connection": self.connection,
            "objectName": object_name,
            "action": action,
            "kwargs": kwargs
        }

        url = f"{self.BACKEND_URL}/api/proxy/"
        return self.call_backend('post', url, json=payload)

    def get(self, object_name, *args, **kwargs):
        kwargs = self.args_to_kwargs(args, kwargs)
        response_dict = self.request_endpoint_via_proxy(object_name, 'get', **kwargs)
        if "detail" in response_dict:
            return response_dict["detail"]
        else:
            return response_dict

    def findone(self, object_name, *args, **kwargs):
        kwargs = self.args_to_kwargs(args, kwargs)
        if "searchterm" not in kwargs:
            raise PeliqanClientException(
                f"Parameter searchterm is required and searchfield is sometimes required for function 'findone'."
            )
        response_dict = self.request_endpoint_via_proxy(object_name, 'findone', **kwargs)
        if "detail" in response_dict:
            detail = response_dict["detail"]
            if not isinstance(detail, dict):  # replace empty string response "" (if no record found) with {}
                detail = {}
            return detail
        else:
            return response_dict

    def list(self, object_name, **kwargs):
        response_dict = self.request_endpoint_via_proxy(object_name, 'list', **kwargs)
        return response_dict

    def add(self, object_name, *args, **kwargs):
        # allow using both keyword arguments or a dict as argument:
        # pq.add("contact", name='John', city='NY') or pq.add("contact", contact_obj)
        kwargs = self.args_to_kwargs(args, kwargs)
        response_dict = self.request_endpoint_via_proxy(object_name, 'add', **kwargs)
        return response_dict

    def update(self, object_name, *args, **kwargs):
        # allow using both keyword arguments or a dict as argument:
        # pq.update("contact", name='John', city='NY') or pq.update("contact", contact_obj)
        kwargs = self.args_to_kwargs(args, kwargs)
        response_dict = self.request_endpoint_via_proxy(object_name, 'update', **kwargs)
        return response_dict

    def upsert(self, object_name, *args, **kwargs):
        # allow using both keyword arguments or a dict as argument:
        # pq.update("contact", name='John', city='NY') or pq.update("contact", contact_obj)
        kwargs = self.args_to_kwargs(args, kwargs)
        if "searchterm" not in kwargs:
            raise PeliqanClientException(
                f"Parameter searchterm is required and searchfield is sometimes required for function 'upsert'.")
        if "searchfield" not in kwargs:
            kwargs["searchfield"] = None
        response_dict_findone = self.request_endpoint_via_proxy(object_name, 'findone', **kwargs)
        kwargs.pop('searchfield', None)
        kwargs.pop('searchterm', None)
        if "detail" in response_dict_findone and "id" in response_dict_findone["detail"]:
            kwargs["id"] = response_dict_findone["detail"]["id"]
            response_dict = self.request_endpoint_via_proxy(object_name, 'update', **kwargs)
        else:
            response_dict = self.request_endpoint_via_proxy(object_name, 'add', **kwargs)
        return response_dict

    def delete(self, object_name, *args, **kwargs):
        kwargs = self.args_to_kwargs(args, kwargs)
        response_dict = self.request_endpoint_via_proxy(object_name, 'delete', **kwargs)
        return response_dict

    def copy(self, object_name, **kwargs):
        response_dict = self.request_endpoint_via_proxy(object_name, 'copy', **kwargs)
        return response_dict

    def rename(self, object_name, **kwargs):
        response_dict = self.request_endpoint_via_proxy(object_name, 'rename', **kwargs)
        return response_dict

    def apicall(self, path, method="get", body=None, **kwargs):
        kwargs["path"] = path
        kwargs["method"] = method
        kwargs["body"] = body
        response_dict = self.request_endpoint_via_proxy("", 'apicall', **kwargs)
        return response_dict



class PeliqanClientException(Exception):
    """
    Base exception raised by the Peliqan module.
    """


class OperationNotSupported(PeliqanClientException):
    """
        Raise this when an operation is not support by the client.
    """


class PeliqanJsonSerializerException(PeliqanClientException):
    """
        Raise this when an json encoding fails for a data structure.
    """



from peliqan.exceptions import PeliqanClientException
from peliqan.client import BaseClient

from peliqan.utils import empty


class BackendServiceClient(BaseClient):
    def __init__(self, jwt, backend_url):
        super(BackendServiceClient, self).__init__(jwt, backend_url)
        self._cache = {
            'connection': {},
            'database': {},
            'schema': {},
            'table': {},
            'interface': {}
        }

    def get_cached_results(self, resource_type, resource_id_or_name, property_name):
        try:
            resource_id_or_name = resource_id_or_name.lower()
        except AttributeError:
            pass

        resource_type = resource_type.lower()
        property_name = property_name.lower()
        return self._cache.get(resource_type, {}).get(resource_id_or_name, {}).get(property_name)

    def _cache_connection_data(self, data):
        connection_name = data.get('connection_name', '')
        if connection_name:
            connection_name = connection_name.lower()
            connection_id = data['connection_id']
            if connection_id:
                data_to_cache = {
                    connection_id: {
                        'connection_name': connection_name
                    },
                    connection_name: {
                        'connection_id': connection_id
                    }
                }
                self._cache['connection'].update(data_to_cache)

    def _cache_database_data(self, data):
        database_name = data.get('database_name', '')
        if database_name:
            database_name = database_name.lower()
            database_id = data['database_id']
            if database_id:
                data_to_cache = {
                    database_id: {
                        'database_name': database_name
                    },
                    database_name: {
                        'database_id': database_id
                    }
                }
                self._cache['database'].update(data_to_cache)

    def _cache_schema_data(self, data):
        schema_name = data.get('schema_name', '')
        if schema_name:
            schema_id = data['schema_id']
            if schema_id:
                data_to_cache = {
                    schema_id: {
                        'schema_name': schema_name
                    },
                    schema_name: {
                        'schema_id': schema_id
                    }
                }
                self._cache['schema'].update(data_to_cache)

    def _cache_table_and_field_data(self, data):
        table_name = data.get('table_name', '')
        if table_name:
            table_id = data['table_id']
            if table_id:
                data_to_cache = {
                    table_name: {
                        'table_id': table_id,
                    },
                    table_id: {
                        'table_name': table_name
                    }
                }
                self._cache['table'].update(data_to_cache)

                # cache field
                field_name = data.get('field_name', '')
                if field_name:
                    field_physical_name = data['field_physical_name']
                    if field_physical_name:
                        data_to_cache[table_name].update({field_name: field_physical_name})
                        data_to_cache[table_id].update({field_name: field_physical_name})

    def _cache_interface_data(self, data):
        interface_name = data.get('interface_name', '')
        if interface_name:
            interface_id = data['interface_id']
            data_to_cache = {
                interface_name: {
                    'interface_id': interface_id,
                },
                interface_id: {
                    'interface_name': interface_name
                }
            }
            self._cache['interface'].update(data_to_cache)

    def _update_cache(self, data):
        self._cache_connection_data(data)
        self._cache_database_data(data)
        self._cache_schema_data(data)
        self._cache_table_and_field_data(data)
        self._cache_interface_data(data)

    def find_connection(self, connection_id=None, connection_name=None):
        """

        :param connection_id:
        :param connection_name:
        :return:
        """
        url = f"{self.BACKEND_URL}/api/database/resource/lookup/connection/"
        data = {
            'connection_id': connection_id,
            'connection_name': connection_name,
        }
        data = self.call_backend("get", url, json=data)

        self._update_cache(data)
        return data

    def find_database(self, connection_id=None, connection_name=None, database_id=None, database_name=None):
        """

        :param connection_id:
        :param connection_name:
        :param database_id:
        :param database_name:
        :return:
        """
        url = f"{self.BACKEND_URL}/api/database/resource/lookup/database/"
        data = {
            'connection_id': connection_id,
            'connection_name': connection_name,
            'database_id': database_id,
            'database_name': database_name,
        }
        data = self.call_backend("get", url, json=data)

        self._update_cache(data)
        return data

    def find_schema(self, connection_id=None, connection_name=None, database_id=None, database_name=None,
                    schema_id=None, schema_name=None):
        """

        :param connection_id:
        :param connection_name:
        :param database_id:
        :param database_name:
        :param schema_id:
        :param schema_name:
        :return:
        """
        url = f"{self.BACKEND_URL}/api/database/resource/lookup/schema/"
        data = {
            'connection_id': connection_id,
            'connection_name': connection_name,
            'database_id': database_id,
            'database_name': database_name,
            'schema_id': schema_id,
            'schema_name': schema_name,
        }
        data = self.call_backend("get", url, json=data)

        self._update_cache(data)
        return data

    def find_table_or_field(
        self, connection_id=None, connection_name=None, database_id=None, database_name=None,
        schema_id=None, schema_name=None, table_id=None, table_name=None,
        field_id=None, field_name=None
    ):
        """
        Find the table information by passing in a table_id or table_name.
        Optionally pass in a field_id or field_name to retrieve that fields info.

        We can also pass in the optional connection_id/connection_name and optional database_id/database_name
        if we want to restrict the search.

        :param connection_id: (Optional) id of the connection the table belongs to.
        :param connection_name: (Optional) name of the connection the table belongs to.
        :param database_id: (Optional) name of the database the table belongs to.
        :param database_name: (Optional) name of the database the table belongs to.
        :param schema_id: (Optional) id of the schema the table belongs to.
        :param schema_name: (Optional) name of the schema the table belongs to.
        :param table_id: id of the table to find.
        :param table_name: name of the table to find.
        :param field_id: (Optional) id of the field to find.
        :param field_name: (Optional) name of the field to find.
        :return:
        """
        url = f"{self.BACKEND_URL}/api/database/resource/lookup/table/"
        data = {
            'connection_id': connection_id,
            'connection_name': connection_name,
            'database_id': database_id,
            'database_name': database_name,
            'schema_id': schema_id,
            'schema_name': schema_name,
            'table_id': table_id,
            'table_name': table_name,
            'field_id': field_id,
            'field_name': field_name
        }
        data = self.call_backend("get", url, json=data)

        self._update_cache(data)
        return data

    def find_interface(self, interface_id=None, interface_name=None):
        url = f"{self.BACKEND_URL}/api/resource/lookup/interface/"
        data = {
            'interface_id': interface_id,
            'interface_name': interface_name
        }
        data = self.call_backend("get", url, json=data)

        self._update_cache(data)
        return data

    def find_resource(self, resource_type, resource_id=None, resource_name=None, **kwargs):

        if resource_type not in ('connection', 'database', 'schema', 'table', 'interface'):
            raise PeliqanClientException(
                f"{resource_type} is not valid. "
                "Allowed resource types are "
                "'connection', 'database', 'schema', 'table', 'script'."
            )

        if not resource_id and not resource_name:
            raise PeliqanClientException("resource_id or resource_name must be provided as kwargs.")

        if resource_type.lower() == 'connection':
            data = {
                'connection_id': resource_id,
                'connection_name': resource_name
            }
            # find connection
            return self.find_connection(**data)

        elif resource_type.lower() == 'database':
            data = {
                'database_id': resource_id,
                'database_name': resource_name,
            }
            # find database
            return self.find_database(**data, **kwargs)

        elif resource_type.lower() == 'schema':
            data = {
                'schema_id': resource_id,
                'schema_name': resource_name
            }
            return self.find_schema(**data, **kwargs)

        elif resource_type.lower() == 'table':
            data = {
                'table_id': resource_id,
                'table_name': resource_name
            }
            return self.find_table_or_field(**data, **kwargs)

        elif resource_type.lower() == 'interface':
            data = {
                'interface_id': resource_id,
                'interface_name': resource_name
            }
            return self.find_interface(**data)

        else:
            raise PeliqanClientException(f"{resource_type} is not valid. "
                                         f"Allowed resource types are 'connection', 'database', 'schema', 'table'.")

    def _is_refresh_allowed(self, resource_type):
        if resource_type not in ['connection', 'database', 'schema', 'table']:
            raise PeliqanClientException(f"{resource_type} is not valid. "
                                         f"Allowed resource types are 'connection', 'database', 'schema', 'table'.")

    def _get_resource_id(self, resource_id, resource_name, resource_type, **kwargs):
        if not resource_id and not resource_name:
            raise PeliqanClientException(f"{resource_type}_id or {resource_type}_name must be provided as kwargs.")

        elif not resource_id and resource_name:
            resource_id = self.get_cached_results(resource_type, resource_name, f'{resource_type}_id')

        if not resource_id:
            lookup_data = self.find_resource(resource_type=resource_type, resource_name=resource_name, **kwargs)
            resource_id = lookup_data[f'{resource_type}_id']

        return resource_id

    def refresh_resource(self, resource_type, refresh_baseurl, resource_name=None, resource_id=None, **kwargs):
        self._is_refresh_allowed(resource_type)
        resource_id = self._get_resource_id(resource_id, resource_name, resource_type, **kwargs)
        url = refresh_baseurl % resource_id

        # call sync url
        response = self.call_backend("get", url, expected_status_code=200)
        prepared_response = {
            'task_id': response['task_id'],
            'detail': response['detail'],
            'syncing': response['syncing']
        }
        return prepared_response

    def get_refresh_resource_task_status(self, resource_type, refresh_baseurl, resource_name=None, resource_id=None,
                                         task_id='', **kwargs):
        self._is_refresh_allowed(resource_type)
        resource_id = self._get_resource_id(resource_id, resource_name, resource_type, **kwargs)
        url = refresh_baseurl % resource_id + f"?task_id={task_id}"

        # call sync task status url
        return self.call_backend('get', url, expected_status_code=200)

    def update_record(self, url, data):
        return self.call_backend("patch", url, json=data, expected_status_code=204)

    def get_cdclogs(self, table_id, writeback_status, change_type, latest_changes_first=False):
        url = f"{self.BACKEND_URL}/api/database/cdclogs/table/{table_id}/?latest_changes_first={latest_changes_first}"

        if writeback_status or change_type:
            url += "&"
            if writeback_status:
                url += f"writeback_status={writeback_status}&"

            if change_type:
                url += f"change_type={change_type}"

        response_dict = self.call_backend("get", url)
        return response_dict

    def update_writeback_status(self, table_id, change_id, writeback_status):
        url = f"{self.BACKEND_URL}/api/database/cdclogs/table/{table_id}/changes/{change_id}/writeback_status/"

        data = {
            "change_id": change_id,
            "writeback_status": writeback_status
        }

        response_dict = self.call_backend("patch", url, json=data)
        return response_dict

    def list_servers(self):
        url = f"{self.BACKEND_URL}/api/servers/"
        response_dict = self.call_backend("get", url)
        return response_dict

    def list_databases(self):
        url = f"{self.BACKEND_URL}/api/applications/"
        response_dict = self.call_backend("get", url)
        return response_dict

    def get_table(self, table_id):
        url = f"{self.BACKEND_URL}/api/database/tables/{table_id}/"
        response_dict = self.call_backend("get", url)
        return response_dict

    def create_query(
        self,
        table_name,
        query,
        database_id=None,
        schema_id=None,
        run_on_peliqan_trino=empty,
        materialize=empty
    ):
        url = f"{self.BACKEND_URL}/api/database/tables/create-sql-query/"
        if schema_id and not database_id:
            response = self.get_schema(schema_id)
            database_id = response['database']

        if database_id:
            url = f"{self.BACKEND_URL}/api/database/tables/database/{database_id}/"

        data = {
            'table_type': 'query',
            'schema_id': schema_id,
            'name': table_name,
            'query': query
        }

        response_dict = self.call_backend("post", url, json=data)

        if run_on_peliqan_trino != empty or materialize != empty:
            response_dict = self.update_table(
                response_dict['id'],
                run_on_peliqan_trino=run_on_peliqan_trino,
                materialize=materialize,
            )

        return response_dict

    def update_table(
        self,
        table_id,
        name=None,
        query=None,
        settings=None,
        run_on_peliqan_trino=empty,
        materialize=empty,
    ):
        url = f"{self.BACKEND_URL}/api/database/tables/%s/" % table_id
        data = {}

        if name:
            data['name'] = name

        if query:
            data['query'] = query

        if run_on_peliqan_trino != empty:
            data['run_on_peliqan_trino'] = run_on_peliqan_trino

        if not settings:
            settings = {}

        if materialize != empty:
            settings['materialize_settings'] = {
                'materialize': materialize
            }

        if settings:
            data['settings'] = settings

        response_dict = self.call_backend("patch", url, json=data)
        return response_dict

    def update_database_metadata(self, database_id, description=None, tags=None):
        url = f"{self.BACKEND_URL}/api/applications/%s/data-catalog/" % database_id
        data = {}
        if description:
            data["description"] = description
        if tags:
            data["tags"] = tags
        response_dict = self.call_backend("patch", url, json=data)
        return response_dict

    def get_schema(self, schema_id):
        url = f"{self.BACKEND_URL}/api/database/schemas/%s/" % schema_id
        response_dict = self.call_backend("get", url)
        return response_dict

    def update_schema(self, schema_id, name):
        url = f"{self.BACKEND_URL}/api/database/schemas/%s/" % schema_id
        data = {}
        if name:
            data["schema_name"] = name
        response_dict = self.call_backend("patch", url, json=data)
        return response_dict

    def update_table_metadata(self, table_id, description=None, tags=None, primary_field_id=None):
        url = f"{self.BACKEND_URL}/api/database/tables/%s/details/" % table_id
        data = {}
        if description:
            data["description"] = description
        if tags:
            data["tags"] = tags

        if primary_field_id:
            data["primary_field_id"] = primary_field_id
        response_dict = self.call_backend("patch", url, json=data)
        return response_dict

    def update_field_metadata(self, field_id, description=None, tags=None):
        url = f"{self.BACKEND_URL}/api/database/fields/%s/data-catalog/" % field_id
        data = {}
        if description:
            data["description"] = description
        if tags:
            data["tags"] = tags
        response_dict = self.call_backend("patch", url, json=data)
        return response_dict

    def list_interfaces(self):
        url = f"{self.BACKEND_URL}/api/interfaces/"
        response_dict = self.call_backend("get", url)
        return response_dict

    def get_interface(self, interface_id):
        url = f"{self.BACKEND_URL}/api/interfaces/{interface_id}/"
        return self.call_backend("get", url)

    def update_interface(
        self,
        interface_id,
        **kwargs
    ):
        url = f"{self.BACKEND_URL}/api/interfaces/{interface_id}/"

        data = {k: v for k, v in kwargs.items() if v is not empty}
        return self.call_backend("patch", url, json=data)

    def create_interface(self, group_id, **kwargs):
        url = f"{self.BACKEND_URL}/api/interfaces/{group_id}/"

        data = {k: v for k, v in kwargs.items() if v is not None}
        return self.call_backend("post", url, json=data)

    def get_interface_state(self, interface_id):
        data = self.get_interface(interface_id)
        return data.get('state', '')

    def set_interface_state(self, interface_id, state):
        url = f"{self.BACKEND_URL}/api/interfaces/{interface_id}/"
        data = {'state': state}
        return self.call_backend("patch", url, json=data)

    def get_pipeline_logs(self, pipeline_run_id):
        if not pipeline_run_id:
            raise PeliqanClientException("'pipeline_run_id' must be provided")

        url = f"{self.BACKEND_URL}/api/pipeline_runs/{pipeline_run_id}/logs"
        return self.call_backend("get", url)

    def get_pipeline_runs(self, connection_id=None, page=1, per_page=10):
        if connection_id:
            url = f"{self.BACKEND_URL}/api/servers/{connection_id}/runs/?page={page}&per_page={per_page}"
        else:
            url = f"{self.BACKEND_URL}/api/pipeline_runs/?page={page}&per_page={per_page}"
        return self.call_backend("get", url)

    def get_interface_runs(self, interface_id=None, page=1, per_page=10):
        url = f"{self.BACKEND_URL}/api/interface_runs/?interface_id={interface_id}&page={page}&per_page={per_page}"
        return self.call_backend("GET", url)

    def send_connection_invite(self, recipient, subject, message, confirmation_recipient, server_type):
        data = {
            'recipient': recipient,
            'subject': subject,
            'message': message,
            'confirmationRecipient': confirmation_recipient,
            'serverType': server_type,
        }

        url = f"{self.BACKEND_URL}/api/servers/invite/"

        return self.call_backend("POST", url, expected_status_code=204, json=data)



__all__ = [
    "BaseClient",
    "WritebackClient",
    "BackendServiceClient",
    "SFTPClient",
    "DBClient",
    "PeliqanTrinoDBClient"
]

from peliqan.client.base import BaseClient
from peliqan.client.writeback import WritebackClient
from peliqan.client.backend_service import BackendServiceClient
from peliqan.client.dbclient import DBClient, PeliqanTrinoDBClient
from peliqan.client.sftpclient import SFTPClient



import requests

from peliqan.exceptions import PeliqanClientException
from peliqan.utils import _serialize_data


class BaseClient:

    def __init__(self, jwt, backend_url):
        self.JWT = jwt
        self.BACKEND_URL = backend_url

    def get_headers(self):
        return {
            "Content-Type": "application/json",
            "Authorization": "JWT %s" % self.JWT
        }

    def call_backend(self, method, url, expected_status_code=200, **kwargs):
        if not kwargs.get('headers'):
            headers = self.get_headers()
            kwargs.update(headers=headers)

        json_data = kwargs.get('json')
        if json_data:
            serialized_data = _serialize_data(json_data)
            kwargs['json'] = serialized_data

        response = requests.request(method, url, **kwargs)
        try:
            response_dict = response.json()
        except ValueError:
            response_dict = {}

        # handle error responses
        if response.status_code != expected_status_code:
            error_message = f"Server responded with status code {response.status_code}"
            if response_dict:
                error_message += f" and error details \n{response_dict}"
            raise PeliqanClientException(error_message)

        return response_dict

    def args_to_kwargs(self, args, kwargs):
        """
        Used to allow using both a dict argument or keyword arguments:
        pq.add("contact", name='John', city='NY') or
        pq.add("contact", contact_obj)
        """
        for arg in args:
            if type(arg) != dict:
                raise PeliqanClientException("Only arguments of type dict and kwargs are accepted")
            kwargs.update(**arg)
        return kwargs



import json


class Streamlit:
    json = json

    def title(self, arg):
        print(arg)

    def text(self, arg):
        print(arg)


st = Streamlit()

RUN_CONTEXT = 'interactive'
from peliqan import Peliqan

pq = Peliqan(
    'eyJ0eXAiOiJKV1QiLCJhbGciOiJIUzI1NiJ9.eyJ1c2VybmFtZSI6InN1cHBvcnQraGV5bG9nK2xrd3dhbHRlckBwZWxpcWFuLmlvIiwiaWF0IjoxNzEwOTM1NTYxLCJleHAiOjI1NzQ4NDkxNjEsImp0aSI6ImEwNTI2ODQ0LWZkYmItNDEzZC04MWJhLWE4MTE0NmZkMjM3MiIsInVzZXJfaWQiOjQ4MSwidXNlcl9wcm9maWxlX2lkIjpbNDc5XSwib3JpZ19pYXQiOjE3MTA5MzU1NjEsInRva2VuX25hbWUiOiJpbnRlcmZhY2VfNDUyIiwiaW50ZXJmYWNlX2lkIjo0NTJ9.1sY9lMdR1j5nj5aA91c-THb5pU4llOGhdv_mCc57HMI',
    'https://app.eu.peliqan.io'
)
pq.INTERFACE_ID = 452

import re
import base64
import io
import pandas as pd

dw = pq.dbconnect('dw_448')
heylog_dev = pq.connect('Heylog dev workspace 2')
heylog_staging = pq.connect('Heylog staging workspace 7')  # Workspace 7 staging is LKWW staging
heylog_production = pq.connect('Heylog production workspace 275')  # Workspace 275 prod is LKWW 24/7 production


def format_lkww_date(lkww_date, lkww_time):
    lkww_date = str(lkww_date)
    lkww_time = str(lkww_time)
    if lkww_time == '0' or int(lkww_time) < 0:  # we had an LIEF_Z_BUC = -41
        lkww_time = '0000'
    iso_date = lkww_date[0:4] + '-' + lkww_date[4:6] + '-' + lkww_date[6:8]
    if len(lkww_time) == 3:
        lkww_time = '0' + lkww_time
    iso_time = lkww_time[0:2] + ':' + lkww_time[2:4] + ':00.000Z'
    return iso_date + 'T' + iso_time


def upsert_order_heylog(lkww_order, heylog_order, env):
    if not lkww_order['source'] == 'JLR_GB_SK.csv':
        # st.text("Skipping order, not from source file JLR_GB_SK.csv")
        return
    if lkww_order['heylog_' + env + '_order_id']:
        # st.text('Updating order in Heylog dev: ' + lkww_order['AG_REFER_POSITION'])
        # heylog_order["id"] = lkww_order['heylog_dev_order_id']
        # result = heylog_dev.update('order_with_labels', heylog_order)

        # To do: Think about updating orders incrementally, avoid updating all orders on each run
        #        Add timestamp changed in lkww_orders table, db.upsert should not increase value if no changes
        #        Or change field status to empty if there is a change
        pass
    else:
        st.text('Adding order to Heylog ' + env + ': ' + lkww_order['AG_REFER_POSITION'])

        if env == 'dev':
            result = heylog_dev.add('order_with_labels', heylog_order)
            if result['status'] == 'error':
                st.json(result)
            else:
                update_status = {
                    'status': result['status'],
                    'heylog_dev_order_id': result['detail'].get('id')
                }
        elif env == 'staging':
            result = heylog_staging.add('order_with_labels', heylog_order)
            if result['status'] == 'error':
                st.json(result)
            else:
                update_status = {
                    'status': result['status'],
                    'heylog_staging_order_id': result['detail'].get('id')
                }
        elif env == 'production':
            result = heylog_production.add('order_with_labels', heylog_order)
            if result['status'] == 'error':
                st.json(result)
            else:
                update_status = {
                    'status': result['status'],
                    'heylog_production_order_id': result['detail'].get('id')
                }
        else:
            st.text("Unknown env in upsert_order_heylog")
            exit()

        if result['status'] == 'success':
            dw.update('dw_448', 'lkww', 'lkww_orders', lkww_order['AG_REFER_POSITION'], update_status)


def transform_lkww_order_to_heylog_order(lkww_order):
    heylog_order = None
    if (lkww_order["AG_REFER"]
            and lkww_order["AG_REFER"] != ""
            and lkww_order["AG_REFER"] != 0
            and lkww_order["AG_REFER"] != "Customer REF (JLR)"):

        licensePlateTruck = lkww_order["KENNZ2"]
        licensePlateTrailer = lkww_order["KENNZ1"]
        firstName = lkww_order["WOB_VORNAM"]
        lastName = lkww_order["WOB_NAME"]
        email = lkww_order["WOB_MAIL"]
        port_from = lkww_order["FAEH_VON_K"]
        port_to = lkww_order["FAEH_NAC_K"]

        # empty cell in Excel becomes NaN, we replace NaN in the DF with 0 (see Excel import code)
        # probably not relevant for CSV files
        if str(licensePlateTruck) == '0':
            licensePlateTruck = ''
        if str(licensePlateTrailer) == '0':
            licensePlateTrailer = ''
        if str(firstName) == '0':
            firstName = ''
        if str(lastName) == '0':
            lastName = ''
        if str(email) == '0':
            email = 'unknown'
        if str(port_from) == '0':
            port_from = ''
        if str(port_to) == '0':
            port_to = ''

        if port_from and port_to:
            port_label = port_from + ' - ' + port_to
        else:
            port_label = ''
        fullName = (firstName + ' ' + lastName).strip()

        heylog_order = {
            "refNumber": lkww_order["AG_REFER"],
            "customer": "JLR",
            "licensePlateTruck": licensePlateTruck,
            "licensePlateTrailer": licensePlateTrailer,
            "start": {
                "name": lkww_order["BELAD_NAME"],
                "street": lkww_order["BELAD_STRA"],
                "zipCode": lkww_order["BELAD_PLZ"],
                "city": lkww_order["BELAD_ORT"]
            },
            "destination": {
                "name": lkww_order["ENTLAD_NAM"],
                "street": lkww_order["ENTLAD_STR"],
                "zipCode": lkww_order["ENTLAD_PLZ"],
                "city": lkww_order["ENTLAD_ORT"]
            },
            "etaFrom": format_lkww_date(lkww_order["LAD_D_BUC"], lkww_order["LAD_Z_BUC"]),
            "etaTo": format_lkww_date(lkww_order["LIEF_D_BUC"], lkww_order["LIEF_Z_BUC"]),
            "labels": [
                {"value": email, "type": "OWNER"},
                {"value": lkww_order["POSITION"], "type": "ORDER_EXTERNAL_REFERENCE"},
            ]
        }
        if port_label:
            heylog_order["labels"].append({"value": port_label})  # generic label, no type

    return heylog_order


def import_orders_from_attachment_into_peliqan_table(email):
    status = "unknown"

    if not "Attachments" in email:
        st.text("Email does not have attachments")
        status = "no attachments"
        return status

    for attachment in email["Attachments"]:
        if not "Content" in attachment and "ContentType" in attachment:
            st.text("Invalid attachment")
            status = "invalid attachment"
            continue
        if attachment["ContentType"] not in ["application/vnd.openxmlformats-officedocument.spreadsheetml.sheet",
                                             "application/octet-stream"]:
            st.text("Skipping attachment, not an Excel or CSV file")
            status = "skipped attachment not Excel or CSV"
            continue

        decoded_bytes = base64.b64decode(attachment["Content"])
        file_bytes = io.BytesIO(decoded_bytes)
        try:
            if ".csv" in attachment["Name"]:
                lkww_orders_df = pd.read_csv(file_bytes, sep='\t', encoding="ISO-8859-1")
                is_csv = True
            else:
                lkww_orders_df = pd.read_excel(file_bytes)
                is_csv = False
        except:
            st.text("Cannot read Excel or CSV file")
            status = "cannot read Excel or CSV file"
            continue

        st.text("Number of rows from file: " + str(len(lkww_orders_df.index)))
        lkww_orders_df = lkww_orders_df.drop(columns=[
            'EXPORT_MRN',
            'T1_MRN',
            'FAEHR_TEXT',
            'FAE_AEND_D',
            'FAE_AEND_Z',
            'FAEHR_PNR',
            'WAB',
            'BAHN_AB_DT',
            'BAHN_AB_ZT',
            'FAEHR_AB_D',
            'ET_ZUGMA',
            'ET_PNR'])
        lkww_orders_df.fillna(0, inplace=True)
        lkww_orders = lkww_orders_df.to_dict('records')

        status = "no orders"
        for lkww_order in lkww_orders:
            lkww_order['AG_REFER_POSITION'] = str(lkww_order['AG_REFER']) + '_' + str(lkww_order['POSITION'])
            if  lkww_order['AG_REFER_POSITION'] == "NT042A-D3-M1/JLR.01568269_SND44945461":
                print(lkww_order, is_csv)
    return status


def process_incoming_webhooks():
    print("Processing incoming webhooks")
    incoming_webhooks = dw.fetch('dw_448', query="SELECT * FROM webhook.incoming_webhooks")
    for incoming_webhook in incoming_webhooks:
        email = incoming_webhook["payload"]
        print("\n\n")
        #
        # extract email when To has this format: "lkww@heylog.com" <lkww@heylog.com>
        if '<' in email["To"] and '>' in email["To"]:
            email["To"] = email["To"][email["To"].find('<') + 1:email["To"].find('>')]

        # Postmark server that receives emails with Excel/CSV attachments (lkww@heylog.com forwards to Postmark)
        if ('lkww@heylog.com' in email["To"].lower() or
                '4f369339798a3c6288eb0a00e84c89e7@inbound.postmarkapp.com' in email["To"].lower()):
            st.text("Processing email: " + email["Subject"] + " To: " + email["To"].lower())
            status = import_orders_from_attachment_into_peliqan_table(email)
            # dw.update('dw_448', 'webhook', 'incoming_webhooks', incoming_webhook["id"], {"status": status})


# def imported_orders_to_heylog():
#     lkww_orders = dw.fetch('dw_448', 'lkww', 'lkww_orders')
#     for lkww_order in lkww_orders:
#         heylog_order = transform_lkww_order_to_heylog_order(lkww_order)
#         if heylog_order:
#             # upsert_order_heylog(lkww_order, heylog_order, 'dev')
#             # upsert_order_heylog(lkww_order, heylog_order, 'staging')
#             upsert_order_heylog(lkww_order, heylog_order, 'production')


process_incoming_webhooks()
st.text('DONE processing incoming webhooks (emails with LKWW orders attachment, from Postmark).')
# imported_orders_to_heylog()
st.text('DONE sending imported LKWW orders to Heylog.')

# ORGINAL CODE

# import json
# import re
# import base64
# import io
# import pandas as pd
#
# dw = pq.dbconnect('dw_448')
# heylog_dev = pq.connect('Heylog dev workspace 2')
# heylog_staging = pq.connect('Heylog staging workspace 7')  # Workspace 7 staging is LKWW staging
# heylog_production = pq.connect('Heylog production workspace 275')  # Workspace 275 prod is LKWW 24/7 production
#
#
# def format_lkww_date(lkww_date, lkww_time):
#     lkww_date = str(lkww_date)
#     lkww_time = str(lkww_time)
#     if lkww_time == '0' or int(lkww_time) < 0:  # we had an LIEF_Z_BUC = -41
#         lkww_time = '0000'
#     iso_date = lkww_date[0:4] + '-' + lkww_date[4:6] + '-' + lkww_date[6:8]
#     if len(lkww_time) == 3:
#         lkww_time = '0' + lkww_time
#     iso_time = lkww_time[0:2] + ':' + lkww_time[2:4] + ':00.000Z'
#     return iso_date + 'T' + iso_time
#
#
# def upsert_order_heylog(lkww_order, heylog_order, env):
#     if not lkww_order['source'] == 'JLR_GB_SK.csv':
#         # st.text("Skipping order, not from source file JLR_GB_SK.csv")
#         return
#     if lkww_order['heylog_' + env + '_order_id']:
#         # st.text('Updating order in Heylog dev: ' + lkww_order['AG_REFER_POSITION'])
#         # heylog_order["id"] = lkww_order['heylog_dev_order_id']
#         # result = heylog_dev.update('order_with_labels', heylog_order)
#
#         # To do: Think about updating orders incrementally, avoid updating all orders on each run
#         #        Add timestamp changed in lkww_orders table, db.upsert should not increase value if no changes
#         #        Or change field status to empty if there is a change
#         pass
#     else:
#         st.text('Adding order to Heylog ' + env + ': ' + lkww_order['AG_REFER_POSITION'])
#
#         if env == 'dev':
#             result = heylog_dev.add('order_with_labels', heylog_order)
#             if result['status'] == 'error':
#                 st.json(result)
#             else:
#                 update_status = {
#                     'status': result['status'],
#                     'heylog_dev_order_id': result['detail'].get('id')
#                 }
#         elif env == 'staging':
#             result = heylog_staging.add('order_with_labels', heylog_order)
#             if result['status'] == 'error':
#                 st.json(result)
#             else:
#                 update_status = {
#                     'status': result['status'],
#                     'heylog_staging_order_id': result['detail'].get('id')
#                 }
#         elif env == 'production':
#             result = heylog_production.add('order_with_labels', heylog_order)
#             if result['status'] == 'error':
#                 st.json(result)
#             else:
#                 update_status = {
#                     'status': result['status'],
#                     'heylog_production_order_id': result['detail'].get('id')
#                 }
#         else:
#             st.text("Unknown env in upsert_order_heylog")
#             exit()
#
#         if result['status'] == 'success':
#             dw.update('dw_448', 'lkww', 'lkww_orders', lkww_order['AG_REFER_POSITION'], update_status)
#
#
# def transform_lkww_order_to_heylog_order(lkww_order):
#     heylog_order = None
#     if (lkww_order["AG_REFER"]
#             and lkww_order["AG_REFER"] != ""
#             and lkww_order["AG_REFER"] != 0
#             and lkww_order["AG_REFER"] != "Customer REF (JLR)"):
#
#         licensePlateTruck = lkww_order["KENNZ2"]
#         licensePlateTrailer = lkww_order["KENNZ1"]
#         firstName = lkww_order["WOB_VORNAM"]
#         lastName = lkww_order["WOB_NAME"]
#         email = lkww_order["WOB_MAIL"]
#         port_from = lkww_order["FAEH_VON_K"]
#         port_to = lkww_order["FAEH_NAC_K"]
#
#         # empty cell in Excel becomes NaN, we replace NaN in the DF with 0 (see Excel import code)
#         # probably not relevant for CSV files
#         if str(licensePlateTruck) == '0':
#             licensePlateTruck = ''
#         if str(licensePlateTrailer) == '0':
#             licensePlateTrailer = ''
#         if str(firstName) == '0':
#             firstName = ''
#         if str(lastName) == '0':
#             lastName = ''
#         if str(email) == '0':
#             email = 'unknown'
#         if str(port_from) == '0':
#             port_from = ''
#         if str(port_to) == '0':
#             port_to = ''
#
#         if port_from and port_to:
#             port_label = port_from + ' - ' + port_to
#         else:
#             port_label = ''
#         fullName = (firstName + ' ' + lastName).strip()
#
#         heylog_order = {
#             "refNumber": lkww_order["AG_REFER"],
#             "customer": "JLR",
#             "licensePlateTruck": licensePlateTruck,
#             "licensePlateTrailer": licensePlateTrailer,
#             "start": {
#                 "name": lkww_order["BELAD_NAME"],
#                 "street": lkww_order["BELAD_STRA"],
#                 "zipCode": lkww_order["BELAD_PLZ"],
#                 "city": lkww_order["BELAD_ORT"]
#             },
#             "destination": {
#                 "name": lkww_order["ENTLAD_NAM"],
#                 "street": lkww_order["ENTLAD_STR"],
#                 "zipCode": lkww_order["ENTLAD_PLZ"],
#                 "city": lkww_order["ENTLAD_ORT"]
#             },
#             "etaFrom": format_lkww_date(lkww_order["LAD_D_BUC"], lkww_order["LAD_Z_BUC"]),
#             "etaTo": format_lkww_date(lkww_order["LIEF_D_BUC"], lkww_order["LIEF_Z_BUC"]),
#             "labels": [
#                 {"value": email, "type": "OWNER"},
#                 {"value": lkww_order["POSITION"], "type": "ORDER_EXTERNAL_REFERENCE"},
#             ]
#         }
#         if port_label:
#             heylog_order["labels"].append({"value": port_label})  # generic label, no type
#
#     return heylog_order
#
#
# def import_orders_from_attachment_into_peliqan_table(email):
#     status = "unknown"
#
#     if not "Attachments" in email:
#         st.text("Email does not have attachments")
#         status = "no attachments"
#         return status
#
#     for attachment in email["Attachments"]:
#         if not "Content" in attachment and "ContentType" in attachment:
#             st.text("Invalid attachment")
#             status = "invalid attachment"
#             continue
#         if attachment["ContentType"] not in ["application/vnd.openxmlformats-officedocument.spreadsheetml.sheet",
#                                              "application/octet-stream"]:
#             st.text("Skipping attachment, not an Excel or CSV file")
#             status = "skipped attachment not Excel or CSV"
#             continue
#
#         decoded_bytes = base64.b64decode(attachment["Content"])
#         file_bytes = io.BytesIO(decoded_bytes)
#         try:
#             if ".csv" in attachment["Name"]:
#                 lkww_orders_df = pd.read_csv(file_bytes, sep='\t', encoding="ISO-8859-1")
#             else:
#                 lkww_orders_df = pd.read_excel(file_bytes)
#         except:
#             st.text("Cannot read Excel or CSV file")
#             status = "cannot read Excel or CSV file"
#             continue
#
#         st.text("Number of rows from file: " + str(len(lkww_orders_df.index)))
#         lkww_orders_df = lkww_orders_df.drop(columns=[
#             'EXPORT_MRN',
#             'T1_MRN',
#             'FAEHR_TEXT',
#             'FAE_AEND_D',
#             'FAE_AEND_Z',
#             'FAEHR_PNR',
#             'WAB',
#             'BAHN_AB_DT',
#             'BAHN_AB_ZT',
#             'FAEHR_AB_D',
#             'ET_ZUGMA',
#             'ET_PNR'])
#         lkww_orders_df.fillna(0, inplace=True)
#         lkww_orders = lkww_orders_df.to_dict('records')
#
#         status = "no orders"
#         for lkww_order in lkww_orders:
#             # AG_REFER is not unique in Excel/CSV files
#             lkww_order['AG_REFER_POSITION'] = str(lkww_order['AG_REFER']) + '_' + str(lkww_order['POSITION'])
#             lkww_order['source'] = attachment["Name"]
#             result = dw.upsert('dw_448', 'lkww', 'lkww_orders', lkww_order['AG_REFER_POSITION'], lkww_order)
#             st.text(lkww_order['AG_REFER_POSITION'] + ': ' + result['status'])
#             status = "processed"
#     return status
#
#
# def process_incoming_webhooks():
#     incoming_webhooks = dw.fetch('dw_448', 'webhook', 'incoming_webhooks')
#     for incoming_webhook in incoming_webhooks:
#         if not incoming_webhook["status"] and incoming_webhook["id"] >= 1163:
#             email = incoming_webhook["payload"]
#
#             # extract email when To has this format: "lkww@heylog.com" <lkww@heylog.com>
#             if '<' in email["To"] and '>' in email["To"]:
#                 email["To"] = email["To"][email["To"].find('<') + 1:email["To"].find('>')]
#
#             # Postmark server that receives emails with Excel/CSV attachments (lkww@heylog.com forwards to Postmark)
#             if ('lkww@heylog.com' in email["To"].lower() or
#                     '4f369339798a3c6288eb0a00e84c89e7@inbound.postmarkapp.com' in email["To"].lower()):
#                 st.text("Processing email: " + email["Subject"] + " To: " + email["To"].lower())
#                 status = import_orders_from_attachment_into_peliqan_table(email)
#                 dw.update('dw_448', 'webhook', 'incoming_webhooks', incoming_webhook["id"], {"status": status})
#
#
# def imported_orders_to_heylog():
#     lkww_orders = dw.fetch('dw_448', 'lkww', 'lkww_orders')
#     for lkww_order in lkww_orders:
#         heylog_order = transform_lkww_order_to_heylog_order(lkww_order)
#         if heylog_order:
#             # upsert_order_heylog(lkww_order, heylog_order, 'dev')
#             # upsert_order_heylog(lkww_order, heylog_order, 'staging')
#             upsert_order_heylog(lkww_order, heylog_order, 'production')
#
#
# process_incoming_webhooks()
# st.text('DONE processing incoming webhooks (emails with LKWW orders attachment, from Postmark).')
# imported_orders_to_heylog()
# st.text('DONE sending imported LKWW orders to Heylog.')



# import necessary modules
import logging
import time

from peliqan.client import WritebackClient, BackendServiceClient, DBClient, SFTPClient, PeliqanTrinoDBClient
from peliqan.exceptions import PeliqanClientException
from peliqan.utils import empty

# get logger
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
logger.propagate = False

# get log handler
sh = logging.StreamHandler()
sh.setLevel(logging.INFO)

# set log format
formatter = logging.Formatter("[%(asctime)s] %(levelname)s %(name)s %(message)s")

# set format to log handler
sh.setFormatter(formatter)

# add handler to logger
logger.addHandler(sh)


class BasePeliqanClient:
    """
    This base class wraps all operations we want to expose to our internal and external clients.
    """

    def __init__(self, jwt, backend_url):
        self.JWT = jwt
        self.BACKEND_URL = backend_url
        self.__service_client__ = BackendServiceClient(jwt, backend_url)

    def find_resource(self, resource_type, resource_id=None, resource_name=None, **kwargs):
        """

           :param resource_type: can be connection/database/table.
           :param resource_id: id of the resource.
           :param resource_name: name of the resource.
           :param kwargs: additional kwargs can also be passed.
           :return: resource details as a dict.
        """
        return self.__service_client__.find_resource(resource_type, resource_id, resource_name, **kwargs)

    # todo allow user to set a PK column for the query (or update the guessed PK)
    def load_table(self, db_name='', schema_name='', table_name='', query='',
                   df=False, fillna_with=None, fillnat_with=None,
                   enable_python_types=True, enable_datetime_as_string=True, tz='UTC'):

        trino_db = self.trinoconnect()
        return trino_db.fetch(db_name=db_name, schema_name=schema_name, table_name=table_name,
                              query=query, df=df,
                              fillna_with=fillna_with, fillnat_with=fillnat_with,
                              enable_python_types=enable_python_types,
                              enable_datetime_as_string=enable_datetime_as_string, tz=tz)

    # todo: make it easier to find table & field ids in UI
    def update_cell(self, row_pk, value, schema_id=None, schema_name=None, table_id=None, table_name=None,
                    field_id=None, field_name=None):
        if row_pk is empty:
            raise PeliqanClientException("'row_pk' must be provided.")

        # BACKEND_URL and JWT are prepended to the generated script,
        # see create_script() --> transform_script() in peliqan/convert_raw_script.py
        base_url = f"{self.BACKEND_URL}/api/database/rows/table/%s/"

        if not schema_id and not schema_name:
            raise PeliqanClientException("schema_id or schema_name must be provided as kwargs")

        if not table_id and not table_name:
            raise PeliqanClientException("table_id or table_name must be provided as kwargs")
        elif not table_id and table_name:
            # only do this if table_id is not provided
            table_id = self.__service_client__.get_cached_results('table', table_name, 'table_id')

        if not table_id:
            lookup_data = self.find_resource(
                'table',
                resource_id=table_id, resource_name=table_name,
                schema_id=schema_id, schema_name=schema_name,
                field_id=field_id, field_name=field_name
            )
            table_id = lookup_data['table_id']

        # # prioritise field id
        if field_id:
            lookup_data = self.find_resource(
                'table',
                resource_id=table_id, resource_name=table_name,
                schema_id=schema_id, schema_name=schema_name,
                field_id=field_id, field_name=field_name
            )
            field_name = lookup_data['table_id']['field_name']

        if not field_name:
            raise PeliqanClientException("field_id or field_name must be provided as kwargs")

        # set the final url
        if isinstance(table_id, int):
            raise PeliqanClientException("table_id could not be resolved, please check provided arguments.")

        # set table_id to base url
        url = base_url % table_id

        data = {
            'pk': row_pk,
            'data': {
                field_name: value
            }
        }
        return self.__service_client__.update_record(url, data)

    def get_refresh_connection_status(self, connection_name=None, connection_id=None, task_id=''):
        base_url = f"{self.BACKEND_URL}/api/servers/%s/syncdb/status/"

        return self.__service_client__.get_refresh_resource_task_status(
            resource_type='connection',
            refresh_baseurl=base_url,
            resource_name=connection_name,
            resource_id=connection_id,
            task_id=task_id
        )

    def get_refresh_database_status(self, connection_name=None, database_name=None, database_id=None, task_id=''):
        base_url = f"{self.BACKEND_URL}/api/applications/%s/syncdb/status/"
        return self.__service_client__.get_refresh_resource_task_status(
            resource_type='database',
            refresh_baseurl=base_url,
            resource_name=database_name,
            resource_id=database_id,
            task_id=task_id,
            connection_name=connection_name
        )

    def get_refresh_schema_status(self, connection_name=None, database_name=None, schema_name=None, schema_id=None,
                                  task_id=''):
        base_url = f"{self.BACKEND_URL}/api/database/schemas/%s/syncdb/status/"
        return self.__service_client__.get_refresh_resource_task_status(
            resource_type='schema',
            refresh_baseurl=base_url,
            resource_name=schema_name,
            resource_id=schema_id,
            task_id=task_id,
            connection_name=connection_name,
            database_name=database_name
        )

    def get_refresh_table_status(self, connection_name=None, database_name=None, schema_name=None, table_name=None,
                                 table_id=None, task_id=''):
        base_url = f"{self.BACKEND_URL}/api/database/tables/%s/syncdb/status/"
        return self.__service_client__.get_refresh_resource_task_status(
            resource_type='table',
            refresh_baseurl=base_url,
            resource_name=table_name,
            resource_id=table_id,
            task_id=task_id,
            connection_name=connection_name,
            database_name=database_name,
            schema_name=schema_name
        )

    def _retry_get_resource_status(self, refresh_func):
        interval = 5  # seconds
        count = 0
        running = True
        while running:
            if count > 10:
                interval = 20

            elif count > 5:
                interval = 10

            response = refresh_func()
            running = response.get('running', True)
            if not running:
                return {
                    'task_id': response['task_id'],
                    'run_id': response.get('run_id'),
                    'detail': 'The sync task has completed.',
                    'syncing': False
                }

            time.sleep(interval)
            count += 1

    def _refresh_connection(self, base_url, connection_id, connection_name, is_async):
        response = self.__service_client__.refresh_resource(
            resource_type='connection',
            refresh_baseurl=base_url,
            resource_name=connection_name,
            resource_id=connection_id
        )

        if not is_async:
            response = self._retry_get_resource_status(
                lambda: self.get_refresh_connection_status(
                    connection_name=connection_name,
                    connection_id=connection_id,
                    task_id=response.get('task_id'))
            )

        return response

    def run_pipeline(self, connection_name=None, connection_id=None, tables=None, is_async=True):
        if not isinstance(tables, (list, tuple, type(None))):
            tables = [tables]

        tables = ','.join(table for table in tables) if tables else ""
        pipeline = "true"

        base_url = (
            f"{self.BACKEND_URL}/api/servers/%s/syncdb/?pipeline={pipeline}&tables={tables}"
        )

        return self._refresh_connection(base_url, connection_id, connection_name, is_async)

    def refresh_connection(
        self,
        connection_name=None,
        connection_id=None,
        is_async=True,
    ):

        base_url = (
            f"{self.BACKEND_URL}/api/servers/%s/syncdb/"
        )

        return self._refresh_connection(base_url, connection_id, connection_name, is_async)

    def refresh_database(self, connection_name=None, database_name=None, database_id=None, is_async=True):
        base_url = f"{self.BACKEND_URL}/api/applications/%s/syncdb/"
        response = self.__service_client__.refresh_resource(resource_type='database', refresh_baseurl=base_url,
                                                            resource_name=database_name, resource_id=database_id,
                                                            connection_name=connection_name)

        if not is_async:
            response = self._retry_get_resource_status(
                lambda: self.get_refresh_database_status(connection_name=connection_name, database_name=database_name,
                                                         database_id=database_id, task_id=response.get('task_id'))
            )

        return response

    def refresh_schema(self, connection_name=None, database_name=None, schema_name=None, schema_id=None, is_async=True):

        base_url = f"{self.BACKEND_URL}/api/database/schemas/%s/syncdb/"
        response = self.__service_client__.refresh_resource(resource_type='schema', refresh_baseurl=base_url,
                                                            resource_name=schema_name, resource_id=schema_id,
                                                            database_name=database_name,
                                                            connection_name=connection_name)

        if not is_async:
            response = self._retry_get_resource_status(
                lambda: self.get_refresh_schema_status(connection_name=connection_name, database_name=database_name,
                                                       schema_name=schema_name, schema_id=schema_id,
                                                       task_id=response.get('task_id'))
            )

        return response

    def refresh_table(self, connection_name=None, database_name=None, schema_name=None,
                      table_name=None, table_id=None, is_async=True):

        base_url = f"{self.BACKEND_URL}/api/database/tables/%s/syncdb/"
        response = self.__service_client__.refresh_resource(resource_type='table', refresh_baseurl=base_url,
                                                            resource_name=table_name, resource_id=table_id,
                                                            database_name=database_name, schema_name=schema_name,
                                                            connection_name=connection_name)

        if not is_async:
            response = self._retry_get_resource_status(
                lambda: self.get_refresh_table_status(connection_name=connection_name, database_name=database_name,
                                                      schema_name=schema_name, table_name=table_name, table_id=table_id,
                                                      task_id=response.get('task_id'))
            )

        return response

    def connect(self, connection=None):
        """
        :param connection: name of the Connection added in Peliqan under Admin > Connections.
        Or a dict with connection properties (credentials etc.).
        """
        if not connection:
            raise PeliqanClientException("connection must be set.")
        connector = WritebackClient(connection, self.JWT, self.BACKEND_URL)
        return connector

    def dbconnect(self, connection=None):
        """
        :param connection: name of the Connection added in Peliqan under Admin > Connections.
        Or a dict with connection properties (credentials etc.).
        """
        if not connection:
            raise PeliqanClientException("connection must be set.")
        connector = DBClient(connection, self.JWT, self.BACKEND_URL)
        return connector

    def trinoconnect(self):
        return PeliqanTrinoDBClient(None, self.JWT, self.BACKEND_URL)

    def sftpconnect(self, connection=None):
        """
        :param connection: name of the Connection added in Peliqan under Admin > Connections.
        Or a dict with connection properties (credentials etc.).
        """
        if not connection:
            raise PeliqanClientException("connection must be set.")
        connector = SFTPClient(connection, self.JWT, self.BACKEND_URL)
        return connector

    def _validate_and_lookup_table(self, table_id, table_name):
        if not table_id and not table_name:
            raise PeliqanClientException("table_id or table_name must be provided as kwargs")
        elif not table_id and table_name:
            # only do this if table_id is not provided
            table_id = self.__service_client__.get_cached_results('table', table_name, 'table_id')

        if not table_id:
            lookup_data = self.find_resource('table', resource_id=table_id, resource_name=table_name)
            table_id = lookup_data['table_id']

        if type(table_id) != int:
            raise PeliqanClientException("table_id could not be resolved, please check provided arguments.")

        return table_id

    def _validate_writeback_status(self, writeback_status):
        error = False
        if type(writeback_status) == list:
            writeback_status_str = ''
            for w in writeback_status:
                w_status = w.upper()
                if w_status not in ["NOT_PROCESSED", "PROCESSED", "CONFIRMED", "FAILED"]:
                    error = True
                    break
                else:
                    writeback_status_str += w_status + ','
            writeback_status = writeback_status_str.rstrip(',')
        elif (
            writeback_status is not None and
            (
                type(writeback_status) != str or writeback_status.upper() not in
                ["NOT_PROCESSED", "PROCESSED", "CONFIRMED", "FAILED"]
            )
        ):
            error = True

        if error:
            raise PeliqanClientException(
                f"writeback_status is not valid. "
                f"Allowed status values are "
                f"'NOT_PROCESSED', 'PROCESSED', 'CONFIRMED', 'FAILED'."
            )

        return writeback_status

    def list_changes(self, table_id=None, table_name=None, writeback_status=None, change_type=None,
                     latest_changes_first=False):
        """
        List the cdc changes in order.
        Optionally, pass writeback_status and/or change_type as a string or list of strings.

        :param table_id: unique integer identifier for a table
        :param table_name: the name or fqn of the table.
        :param writeback_status: valid string or list of string values
        :param change_type: valid string or list of string values
        :param latest_changes_first: use this to get toggle the order of changes (Asc/Desc of id). default is False.
        :return:
        """

        error = False
        # i = insert, u = update, d = delete, t = transformation, f = formula, l = link (to another table)
        # m = multiselect
        if type(change_type) == list:
            change_type_str = ''
            for c in change_type:
                c_type = c.lower()
                if c_type not in ["i", "u", "d", "f", "t", "m", "l"]:
                    error = True
                    break
                else:
                    change_type_str += c_type + ','

            change_type = change_type_str.rstrip(',')
        elif (
            change_type is not None and
            (type(change_type) != str or change_type.lower() not in ["i", "u", "d", "f", "t", "m", "l"])
        ):
            error = True

        if error:
            raise PeliqanClientException(
                f"change_type is not valid. Allowed status values are 'i', 'u', 'd', 'f', 't', 'm', 'l'.\n"
                f"i = insert, u = update, d = delete, t = transformation, f = formula, l = link (to another table)"
                f"m = multiselect."
            )

        writeback_status = self._validate_writeback_status(writeback_status)

        table_id = self._validate_and_lookup_table(table_id, table_name)

        return self.__service_client__.get_cdclogs(table_id=table_id, writeback_status=writeback_status,
                                                   change_type=change_type, latest_changes_first=latest_changes_first)

    def update_writeback_status(self, change_id, writeback_status, table_id=None, table_name=None):
        """
        Update the writeback_status for a cdc log.

        :param change_id: unique integer identifier for a cdc log.
        :param writeback_status: valid status value.
        :param table_id: unique integer identifier for a table
        :param table_name: the name or fqn of the table.
        :return:
        """
        table_id = self._validate_and_lookup_table(table_id, table_name)
        if (
            type(writeback_status) != str or
            writeback_status.upper() not in ["NOT_PROCESSED", "PROCESSED", "CONFIRMED", "FAILED"]
        ):
            raise PeliqanClientException(
                f"writeback_status is not valid. "
                f"Allowed status values are "
                f"'NOT_PROCESSED', 'PROCESSED', 'CONFIRMED', 'FAILED'."
            )

        try:
            change_id = int(change_id)
        except ValueError:
            raise PeliqanClientException("change_id must be a valid integer")

        return self.__service_client__.update_writeback_status(table_id, change_id, writeback_status)

    def list_connections(self):
        """
        Returns a list of servers in the account.

        :return: list of databases
        """
        return self.__service_client__.list_servers()

    def list_databases(self):
        """
        Returns a list of all databases in the account including tables and fields in tables.

        :return: list of databases
        """
        return self.__service_client__.list_databases()

    def get_schema(self, schema_id):
        """
        Returns all meta-data for a schema.

        :return: schema meta-data
        """
        return self.__service_client__.get_schema(schema_id)

    def get_table(self, table_id):
        """
        Returns all meta-data for a table including fields.

        :return: table meta-data
        """
        return self.__service_client__.get_table(table_id)

    def update_field(self, field_id, description=None, tags=None):
        """
        Updates a field (column).

        :param field_id: required, integer, id of the field to update
        :param description: optional, string, description of the field (data catalog metadata)
        :param tags: optional, array of strings, tags assigned to the field (data catalog metadata)
        :return: result of update
        """

        return self.__service_client__.update_field_metadata(field_id, description, tags)

    def update_table(
        self,
        table_id,
        name=None,
        query=None,
        primary_field_id=None,
        settings=None,
        description=None,
        tags=None,
        run_on_peliqan_trino=empty,
        materialize=empty
    ):
        """
        Updates a table.

        :param table_id: required, integer, id of the table to update
        :param name: optional, string, new name of the table
        :param query: optional, string, new SQL query for tables of type 'query'
        :param primary_field_id: optional, integer, primary key field id for this table, i.e. the field id in Peliqan.
        See, table details page in Peliqan to get the primary_field_id for a table.
        :param settings: optional, string (json), settings of the table
        :param description: optional, string, description of the table (data catalog metadata)
        :param tags: optional, array of strings, tags assigned to the table (data catalog metadata)
        :param run_on_peliqan_trino: optional, boolean, whether the query should run on the native engine or trino engine.
        :param materialize: optional, boolean,  whether the query should be materialized as a table into a database.

        :return: result of update
        """

        update_result_dict = {}
        update_metadata_result_dict = {}
        if name or query or settings or run_on_peliqan_trino != empty or materialize != empty:
            update_result_dict = self.__service_client__.update_table(
                table_id,
                name,
                query,
                settings,
                run_on_peliqan_trino,
                materialize
            )
        if description or tags or primary_field_id:
            update_metadata_result_dict = self.__service_client__.update_table_metadata(
                table_id,
                description,
                tags,
                primary_field_id
            )
        return {**update_result_dict, **update_metadata_result_dict}

    def update_schema(self, schema_id, name):
        """
        Updates a schema.

        :param schema_id: required, integer, id of the schema to update
        :param name: required, new name of the schema
        :return: result of update
        """

        return self.__service_client__.update_schema(schema_id, name)

    def update_database(self, database_id, description=None, tags=None):
        """
        Updates a database.

        :param database_id: required, integer, id of the database to update
        :param description: optional, string, description of the database (data catalog metadata)
        :param tags: optional, array of strings, tags assigned to the database (data catalog metadata)
        :return: result of update
        """

        return self.__service_client__.update_database_metadata(database_id, description, tags)

    def list_scripts(self):
        return self.__service_client__.list_interfaces()

    def get_script(self, script_id=None, script_name=None):
        if not script_id and not script_name:
            raise PeliqanClientException("'script_id' or 'script_name' must be provided")

        if not script_id and script_name:
            script_id = self.__service_client__.get_cached_results('interface', script_name, 'interface_id')

        if not script_id:
            interface_data = self.__service_client__.find_resource('interface', resource_name=script_name)
            script_id = interface_data['interface_id']

        return self.__service_client__.get_interface(script_id)

    def update_script(
        self,
        script_id,
        name=empty,
        group=empty,
        raw_script=empty,
        settings=empty,
        state=empty,
        flow=empty,
        editor=empty
    ):
        """
        A function to update a script in Peliqan.

        :param script_id: The id of the script that needs to be updated.
        :param name: A string value that represents the new name of the script.
        :param group: The new group id that the script should belong to.
        :param raw_script: The python code that should be associated with this script.
        :param settings: Update the script run schedule settings.
        :param state: Update the state of the script.
        :param flow: The json settings associated with the visual flow editor.
        :param editor: An enum value that decides which editor type must be opened in the Peliqan code editor.
        Options: [RAW_SCRIPT_EDITOR, FLOW_EDITOR]
        :return:
        """
        return self.__service_client__.update_interface(
            script_id,
            name=name,
            group=group,
            raw_script=raw_script,
            settings=settings,
            state=state,
            flow=flow,
            editor=editor
        )

    def add_script(self, group_id, name, run_mode='STREAMLIT', script_type='streamlit'):
        """

        :param group_id: The group the script will belong to.
        :param name: The name of the new script.
        :param script_type: An enum value that decides the type of script. Options: [streamlit]
        :param run_mode:  An enum value that decides whether the script will be triggered by an API or a streamlit app.
        Options: [STREAMLIT, API, SHELL]
        :return:
        """
        return self.__service_client__.create_interface(
            group_id,
            name=name,
            type=script_type,
            run_mode=run_mode
        )

    def add_query_table(
        self,
        table_name,
        query,
        database_id=None,
        schema_id=None,
        run_on_peliqan_trino=empty,
        materialize=empty,
    ):
        return self.__service_client__.create_query(
            table_name=table_name,
            query=query,
            database_id=database_id,
            schema_id=schema_id,
            run_on_peliqan_trino=run_on_peliqan_trino,
            materialize=materialize
        )

    def generate_sql_union(self, table_ids, sources=None):
        """
        Generates an SQL UNION query for the given tables.
        All columns of all tables will be added to the UNION query.
        If a column does not exist in one of the tables, it will be added with a null value.
        Optionally, a "source" column can be added to indicate from which table each row originated.

        :param table_ids: required, list of integers, list of table ids to include in UNION query
        :param sources: optional, dict, if set an extra 'source' column will be added to indicate to which table the record belongs. Keys are table ids. Values are source value to include in UNION result. Example: { 1: "Paris", 2: "London" }. This will add a column "source" to the UNION where all records from table id 1 will have value "Paris" for the source.
        :return: result of update
        """

        tables = []
        fields = []
        field_types = {}
        fields_to_cast = []
        for table_id in table_ids:
            table = self.get_table(table_id)
            tables.append(table)
            for field in table["all_fields"]:
                if field["name"] not in fields:
                    fields.append(field["name"])
                    if field["sql_data_type"]:
                        # Actual field type in source, e.g. timestamp, timestamptz...
                        # (might not be available for all sources)
                        field_types[field["name"]] = field["sql_data_type"]
                    else:
                        # Peliqan field type, e.g. "date"
                        field_types[field["name"]] = field["type"]
                elif (
                    field["sql_data_type"] and field["sql_data_type"] != field_types[field["name"]]
                ) or field["type"] != field_types[field["name"]]:
                    fields_to_cast.append(field["name"])

        table_selects = []
        for table in tables:
            table_select_fields = []
            if sources:
                source_name = ""
                for source_table_id, source_table_name in sources.items():
                    if int(source_table_id) == table['id']:
                        source_name = source_table_name
                table_select_fields.append("'%s'" % source_name + " source")
            for field in fields:
                table_has_field = False
                for table_field in table["all_fields"]:
                    if (
                        field[0] == '"' and table_field["name"] == field) or (
                        field[0] != '"' and table_field["name"].lower() == field.lower()
                    ):
                        table_has_field = True
                        break
                if table_has_field:
                    if field in fields_to_cast:
                        if table_field["sql_data_type"] and table_field["sql_data_type"] == "timestamptz":
                            # Postgres fields of type timestamptz (timestamp with timezone)
                            # cannot be cast to varchar directly by Trino
                            table_select_fields.append("CAST(CAST(%s AS TIMESTAMP) AS VARCHAR) AS %s" % (field, field))
                        else:
                            table_select_fields.append("CAST(%s AS VARCHAR) AS %s" % (field, field))
                    else:
                        table_select_fields.append(field)
                else:
                    table_select_fields.append("null " + field)
            table_select_fields_str = ", ".join(table_select_fields)

            table_select = "SELECT %s FROM %s" % (table_select_fields_str, table["name"])
            table_selects.append(table_select)

        union = " UNION ALL ".join(table_selects)
        return union

    def get_interface_state(self, interface_id):
        """
        An interface is a saved program. Get the stored state for a specific interface.

        :param interface_id: The id of the interface.
        :return: Any
        """
        return self.__service_client__.get_interface_state(interface_id)

    def set_interface_state(self, interface_id, state):
        """
        An interface is a saved program. Set the state for a specific interface in the peliqan environment.

        :param interface_id: The id of the interface.
        :param state: The data that will be stored as the state value for an interface.
        :return:
        """
        return self.__service_client__.set_interface_state(interface_id, state)

    def get_pipeline_logs(self, pipeline_run_id):
        """
        Retrieves the pipeline logs of a SaaS connection.

        :param pipeline_run_id: id of the pipeline run.

        :return: logs of the pipeline
        """
        return self.__service_client__.get_pipeline_logs(pipeline_run_id)

    def get_pipeline_runs(self, connection_id=None, page=1, per_page=10):
        """
        Retrieves the pipeline logs of a SaaS connection.

        :param connection_id: integer, id of the source connection.
        :param page: integer, the page number to fetch.
        :param per_page: integer, the number of results per page.
        :return: list of pipeline runs
        """
        return self.__service_client__.get_pipeline_runs(connection_id, page, per_page)

    def get_script_runs(self, script_id=None, page=1, per_page=10):
        return self.__service_client__.get_interface_runs(script_id, page, per_page)

    def send_connection_invite(
        self,
        recipient,
        subject,
        message,
        confirmation_recipient,
        server_type
    ):

        return self.__service_client__.send_connection_invite(
            recipient,
            subject,
            message,
            confirmation_recipient,
            server_type
        )

    def get_or_create_schema(self, schema_name, database_name=None, database_id=None):
        if not database_id and not database_name:
            raise PeliqanClientException("'database_name' or 'database_id' must be provided.")

        if not database_id and database_name:
            response = self.find_resource('database', resource_name=database_name)
            database_id = response['database_id']

        try:
            url = f"{self.BACKEND_URL}/api/database/schemas/database/{database_id}/"
            response = self.__service_client__.call_backend("POST", url, json={'schema_name': schema_name})
        except PeliqanClientException as e:
            if 'ERROR_SCHEMA_ALREADY_EXISTS' not in str(e):
                raise

            response = self.__service_client__.find_schema(database_id=database_id, schema_name=schema_name)
            schema_id = response.get('schema_id')
            if schema_id:
                response = self.get_schema(schema_id)

        return response

    def upsert_query_table(
        self,
        table_name,
        query,
        schema_id=None,
        schema_name=None,
        database_id=None,
        database_name=None,
        run_on_peliqan_trino=empty,
        materialize=empty
    ):

        if not database_id and not database_name and not schema_id and not schema_name:
            schema_name = "My Queries"  # default `My Queries` schema in database
            # todo: database_id = self.DW_ID

        try:
            response = self.__service_client__.find_table_or_field(
                database_id=database_id,
                database_name=database_name,
                schema_id=schema_id,
                schema_name=schema_name,
                table_name=table_name
            )

            table_id = response.get('table_id')
            if table_id:
                response = self.update_table(
                    table_id,
                    name=table_name,
                    query=query,
                    run_on_peliqan_trino=run_on_peliqan_trino,
                    materialize=materialize
                )

        except PeliqanClientException as e:
            if 'ERROR_TABLE_DOES_NOT_EXIST' not in str(e):
                raise

            if not database_id and database_name:
                try:
                    response = self.find_resource('database', resource_name=database_name)
                    database_id = response['database_id']
                except PeliqanClientException as e:
                    if (
                        'ERROR_APPLICATION_DOES_NOT_EXIST' not in str(e) or
                        "ERROR_DATABASE_DOES_NOT_EXIST" not in str(e)
                    ):
                        raise

            if not schema_id and schema_name:
                try:
                    response = self.find_resource(
                        'schema',
                        resource_name=schema_name,
                        database_name=database_name,
                        database_id=database_id
                    )
                    schema_id = response['schema_id']

                except PeliqanClientException as e:
                    if 'ERROR_SCHEMA_DOES_NOT_EXIST' not in str(e):
                        raise

                    # if not the default schema then perform the schema retrieval
                    if schema_name != 'My Queries':
                        # todo: database_id = database_id or self.DW_ID
                        response = self.get_or_create_schema(schema_name, database_id=database_id)
                        schema_id = response['id']

            response = self.add_query_table(
                table_name,
                query,
                database_id=database_id,
                schema_id=schema_id,
                run_on_peliqan_trino=run_on_peliqan_trino,
                materialize=materialize,
            )

        return response

    def discover_object_schema(self, records):
        if not isinstance(records, (list, tuple)):
            records = [records]

        url = f"{self.BACKEND_URL}/api/database/resource/json-schema/"
        return self.__service_client__.call_backend("POST", url, json={'records': records})

